/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db.base;

import ch.ehi.basics.logging.EhiLogger;


/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 15.03.2007 $
 */
public class DbIdGen {
	public final static String SQL_T_KEY_OBJECT="T_Key_Object";
	public final static String SQL_T_KEY="T_Key";
	public final static String SQL_T_LASTUNIQUEID="T_LastUniqueId";
	java.sql.Connection conn=null;
	String dbusr=null;
	public DbIdGen(java.sql.Connection conn,String dbusr)
	{
		this.conn=conn;
		this.dbusr=dbusr;
	}
//	CREATE TABLE T_Key_Object (
//		   T_Key                Text(30) NOT NULL,
//		   T_LastUniqueId       LONG NOT NULL,
//		   T_CreateDate         DATE NOT NULL,
//		   T_LastChange         DATE NOT NULL,
//		   T_User               TEXT(30) NOT NULL,
//		   CONSTRAINT PK PRIMARY KEY (T_Key)
//	);
	 
	static public void addMappingTable(ch.ehi.sqlgen.repository.DbSchema schema)
	{
		ch.ehi.sqlgen.repository.DbTable tab=new ch.ehi.sqlgen.repository.DbTable();
		tab.setName(SQL_T_KEY_OBJECT);
		ch.ehi.sqlgen.repository.DbColVarchar t_key=new ch.ehi.sqlgen.repository.DbColVarchar();
		t_key.setName(SQL_T_KEY);
		t_key.setNotNull(true);
		t_key.setSize(30);
		tab.addColumn(t_key);
		ch.ehi.sqlgen.repository.DbColNumber t_lastuniqueid=new ch.ehi.sqlgen.repository.DbColNumber();
		t_lastuniqueid.setName(SQL_T_LASTUNIQUEID);
		t_lastuniqueid.setNotNull(true);
		tab.addColumn(t_lastuniqueid);
		ch.ehi.ili2db.fromili.TransferFromIli.addStdCol(tab);
		schema.addTable(tab);
	}
	public int getLastId(String key)
	{
		java.sql.PreparedStatement getstmt=null;
		try{
			getstmt=conn.prepareStatement("SELECT "+SQL_T_LASTUNIQUEID+" FROM "+SQL_T_KEY_OBJECT+" WHERE "+SQL_T_KEY+"= ?");
			getstmt.setString(1,key);
			java.sql.ResultSet res=getstmt.executeQuery();
			int ret=0;
			if(res.next()){
				ret=res.getInt(1);
				return ret;
			}
		}catch(java.sql.SQLException ex){
			EhiLogger.logError("failed to query "+SQL_T_KEY_OBJECT,ex);
		}finally{
			if(getstmt!=null){
				try{
					getstmt.close();
				}catch(java.sql.SQLException ex){
					EhiLogger.logError(ex);
				}
			}
		}
			
		// INSERT
		int ret=0;
		java.sql.Timestamp today=new java.sql.Timestamp(System.currentTimeMillis());
		String stmt="INSERT INTO "+SQL_T_KEY_OBJECT+" ("+SQL_T_KEY+","+SQL_T_LASTUNIQUEID+",T_LastChange,T_CreateDate,T_User) VALUES (?,?,?,?,?)";
		java.sql.PreparedStatement updstmt=null;
		try{
			updstmt = conn.prepareStatement(stmt);
			updstmt.setString(1, key);
			updstmt.setInt(2, ret);
			updstmt.setTimestamp(3, today);
			updstmt.setTimestamp(4, today);
			updstmt.setString(5, dbusr);
			updstmt.executeUpdate();
		}catch(java.sql.SQLException ex){
			EhiLogger.logError("failed to insert "+SQL_T_KEY_OBJECT,ex);
		}finally{
			if(updstmt!=null){
				try{
					updstmt.close();
				}catch(java.sql.SQLException ex){
					EhiLogger.logError(ex);
				}
			}
		}
		return ret;
	}
	public void setLastId(int newId,String key)
	{
		// update entry
		java.sql.Timestamp today=new java.sql.Timestamp(System.currentTimeMillis());
		String stmt="UPDATE "+SQL_T_KEY_OBJECT+" SET "+SQL_T_LASTUNIQUEID+"=?,T_LastChange=?,T_User=? WHERE "+SQL_T_KEY+"=?";
		java.sql.PreparedStatement updstmt = null;
		try{
			updstmt = conn.prepareStatement(stmt);
			updstmt.setInt(1, newId);
			updstmt.setTimestamp(2, today);
			updstmt.setString(3, dbusr);
			updstmt.setString(4, key);
			updstmt.executeUpdate();
		}catch(java.sql.SQLException ex){
			EhiLogger.logError("failed to update "+SQL_T_KEY_OBJECT,ex);
		}finally{
			if(updstmt!=null){
				try{
					updstmt.close();
				}catch(java.sql.SQLException ex){
					EhiLogger.logError(ex);
				}
			}
		}		
	}
}
