/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db.base;

import ch.ehi.basics.logging.EhiLogger;
import ch.interlis.ili2c.metamodel.TransferDescription;
import ch.interlis.ili2c.Ili2cFailure;
import ch.interlis.ili2c.ModelScan;
import ch.interlis.ili2c.config.Configuration;
import ch.ehi.sqlgen.repository.DbSchema;
import ch.ehi.sqlgen.generator.*;
//import ch.ehi.sqlgen.generator_impl.oracle.GeneratorOracle;
import ch.ehi.sqlgen.generator_impl.jdbc.*;
import ch.ehi.sqlgen.generator.Generator;
import ch.ehi.ili2db.converter.*;
import ch.ehi.ili2db.fromili.*;
import ch.ehi.ili2db.fromxtf.*;
import ch.ehi.ili2db.mapping.*;
import ch.ehi.ili2db.toxtf.*;
//import ch.interlis.iom.swig.iom_javaConstants;
import ch.interlis.iom_j.xtf.XtfWriter;
import ch.interlis.iom_j.xtf.XtfReader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;

import java.sql.Connection;
import java.sql.DriverManager;
import ch.ehi.ili2db.gui.Config;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 07.02.2005 $
 */
public class Ili2db {
	public static final String XTF="%XTF";
	public static final String XTF_DIR="%XTF_DIR";
	public static final String JAR_DIR="%JAR_DIR";
	public static void run(Config config,String appHome)
	throws Ili2dbException
	{
		if(config.getFunction()==Config.FC_IMPORT){
			runImport(config,appHome);
		}else if(config.getFunction()==Config.FC_EXPORT){
			runExport(config,appHome);
		}else if(config.getFunction()==Config.FC_SCHEMAIMPORT){
			runSchemaImport(config,appHome);
		}else{
			EhiLogger.logError("function not supported");
			return;
		}
	}
	public static void runImport(Config config,String appHome) 
	throws Ili2dbException
		{
		String xtffile=config.getXtffile();
			if(xtffile==null){
				EhiLogger.logError("no xtf-file given");
				return;
			}
			String modeldir=config.getModeldir();
			if(modeldir==null){
				EhiLogger.logError("no modeldir given");
				return;
			}
		ArrayList modeldirv = getModelDirv(modeldir, xtffile,appHome);
			
		String models=config.getModels();
		ArrayList modelv=new ArrayList();
		if(models==null){
			EhiLogger.logError("no models given");
			return;
		}
		String modelnames[]=models.split(";");
		for(int modeli=0;modeli<modelnames.length;modeli++){
			String m=modelnames[modeli];
			if(m.equals(XTF)){
				// read modelname from xtf-file
				m=getModelFromXtf(xtffile);
			}
			if(m!=null){
				modelv.add(m);				
			}
		}
		if(modelv.size()==0){
			EhiLogger.logError("no models given");
			return;
		}

		String dburl=config.getDburl();
		String dbusr=config.getDbusr();
		String dbpwd=config.getDbpwd();
		if(dburl==null){
			EhiLogger.logError("no dburl given");
			return;
		}
		if(dbusr==null){
			//EhiLogger.logError("no dbusr given");
			//return;
			dbusr="";
		}
		if(dbpwd==null){
			//EhiLogger.logError("no dbpwd given");
			//return;
			dbpwd="";
		}
		String geometryConverter=config.getGeometryConverter();
		if(geometryConverter==null){
			EhiLogger.logError("no geoemtry converter given");
			return;
		}
		String ddlGenerator=config.getDdlGenerator();
		if(ddlGenerator==null){
			EhiLogger.logError("no DDL generator given");
			return;
		}
		String jdbcDriver=config.getJdbcDriver();
		if(jdbcDriver==null){
			EhiLogger.logError("no JDBC driver given");
			return;
		}
		
		// compile required ili files

		EhiLogger.logState("compile models...");
		TransferDescription td;
		try {
			if(((String)modelv.get(0)).endsWith(".ili")){
				td = ch.interlis.ili2c.Ili2c.compileIliFiles(modelv,modeldirv);
			}else{
				td = ch.interlis.ili2c.Ili2c.compileIliModels(modelv,modeldirv);				
			}
		} catch (Ili2cFailure ex) {
			EhiLogger.logError(ex);
			return;
		}

		// open db connection
		try{
			Class.forName(jdbcDriver);
		}catch(Exception ex){
			EhiLogger.logError("failed to load JDBC driver",ex);
			return;
		}
		Connection conn=null;
		String url = dburl;
		try{
		  //DriverManager.registerDriver(new oracle.jdbc.OracleDriver());
		  conn = DriverManager.getConnection(url, dbusr, dbpwd);
		}catch(Exception ex){
			EhiLogger.logError(ex);
			return;
		}

		// read mapping file
		String mappingConfig=config.getMappingConfig();
		Mapping mapping=new Mapping();
		if(mappingConfig!=null){
			mapping.readConfig(mappingConfig);
		}

		Generator gen=null;
		try{
			gen=(Generator)Class.forName(ddlGenerator).newInstance();
		}catch(Exception ex){
			EhiLogger.logError("failed to load/create DDL generator",ex);
			return;
		}
		SqlGeometryConverter geomConverter=null;
		try{
			geomConverter=(SqlGeometryConverter)Class.forName(geometryConverter).newInstance();
		}catch(Exception ex){
			EhiLogger.logError("failed to load/create geometry converter",ex);
			return;
		}
		geomConverter.setup(conn, config);

		// create table structure
		EhiLogger.logState("create table structure...");
		try{
			TransferFromIli trsfFromIli=new TransferFromIli();
			
			// map ili-classes to sql-tables
			DbSchema schema;
			try {
				schema = trsfFromIli.doit(td,mapping,config.getDefaultSrsAuthority(),config.getDefaultSrsCode());
			} catch (Ili2dbException e) {
				EhiLogger.logError("mapping of ili-classes to sql-tables failed",e);
				return;
			}
			if(schema==null){
				return;
			}

			TransferFromIli.addInheritanceTable(schema);
			Mapping.addMappingTable(schema);
			DbIdGen.addMappingTable(schema);
						
			GeneratorDriver drv=new GeneratorDriver(gen);
			SqlConfig sqlconfig=new SqlConfig();
			sqlconfig.setJdbcConnection(conn);
			drv.visitSchema(sqlconfig,schema);
			// create script requested by user?
			String createscript=config.getCreatescript();
			if(createscript!=null && (gen instanceof GeneratorJdbc)){
				writeScript(createscript,((GeneratorJdbc)gen).iteratorCreateLines());
			}
			// drop script requested by user?
			String dropscript=config.getDropscript();
			if(dropscript!=null && (gen instanceof GeneratorJdbc)){
				writeScript(dropscript,((GeneratorJdbc)gen).iteratorDropLines());
			}
			// update mapping table
			mapping.updateMappingTable(conn);
			// update inheritance table
			trsfFromIli.updateInheritanceTable(conn);
		}catch(java.io.IOException ex){
			EhiLogger.logError(ex);
			return;
		}
		
		// process xtf files
		EhiLogger.logState("process data file...");
		transferFromXtf(conn,xtffile,mapping,td,dbusr,geomConverter,config.getDefaultSrsAuthority(),config.getDefaultSrsCode());
		try{
			conn.close();
			EhiLogger.logState("...import done");
		}catch(java.sql.SQLException ex){
			EhiLogger.logError(ex);
		}
	}
	public static void runSchemaImport(Config config,String appHome) 
	throws Ili2dbException
	{
		String modeldir=config.getModeldir();
		if(modeldir==null){
			EhiLogger.logError("no modeldir given");
			return;
		}
	ArrayList modeldirv = getModelDirv(modeldir, null,appHome);
		
	String models=config.getModels();
	ArrayList modelv=new ArrayList();
	if(models==null){
		EhiLogger.logError("no models given");
		return;
	}
	String modelnames[]=models.split(";");
	for(int modeli=0;modeli<modelnames.length;modeli++){
		String m=modelnames[modeli];
		if(m!=null && m.equals(XTF)){
			// read modelname from xtf-file
			EhiLogger.logError("illegal use of parameter "+XTF);
			return;
		}
		if(m!=null){
			modelv.add(m);				
		}
	}
	if(modelv.size()==0){
		EhiLogger.logError("no models given");
		return;
	}

	String dburl=config.getDburl();
	String dbusr=config.getDbusr();
	String dbpwd=config.getDbpwd();
	if(dburl==null){
		EhiLogger.logError("no dburl given");
		return;
	}
	if(dbusr==null){
		//EhiLogger.logError("no dbusr given");
		//return;
		dbusr="";
	}
	if(dbpwd==null){
		//EhiLogger.logError("no dbpwd given");
		//return;
		dbpwd="";
	}
	String geometryConverter=config.getGeometryConverter();
	if(geometryConverter==null){
		EhiLogger.logError("no geoemtry converter given");
		return;
	}
	String ddlGenerator=config.getDdlGenerator();
	if(ddlGenerator==null){
		EhiLogger.logError("no DDL generator given");
		return;
	}
	String jdbcDriver=config.getJdbcDriver();
	if(jdbcDriver==null){
		EhiLogger.logError("no JDBC driver given");
		return;
	}
	
	// compile required ili files

	EhiLogger.logState("compile models...");
	TransferDescription td;
	try {
		if(((String)modelv.get(0)).endsWith(".ili")){
			td = ch.interlis.ili2c.Ili2c.compileIliFiles(modelv,modeldirv);
		}else{
			td = ch.interlis.ili2c.Ili2c.compileIliModels(modelv,modeldirv);				
		}
	} catch (Ili2cFailure ex) {
		EhiLogger.logError(ex);
		return;
	}
	// open db connection
	try{
		Class.forName(jdbcDriver);
	}catch(Exception ex){
		EhiLogger.logError("failed to load JDBC driver",ex);
		return;
	}
	Connection conn=null;
	String url = dburl;
	try{
	  conn = DriverManager.getConnection(url, dbusr, dbpwd);
	}catch(Exception ex){
		EhiLogger.logError(ex);
		return;
	}

	// read mapping file
	String mappingConfig=config.getMappingConfig();
	Mapping mapping=new Mapping();
	if(mappingConfig!=null){
		mapping.readConfig(mappingConfig);
	}

	Generator gen=null;
	try{
		gen=(Generator)Class.forName(ddlGenerator).newInstance();
	}catch(Exception ex){
		EhiLogger.logError("failed to load/create DDL generator",ex);
		return;
	}
	SqlGeometryConverter geomConverter=null;
	try{
		geomConverter=(SqlGeometryConverter)Class.forName(geometryConverter).newInstance();
	}catch(Exception ex){
		EhiLogger.logError("failed to load/create geometry converter",ex);
		return;
	}
	geomConverter.setup(conn, config);

	// create table structure
	EhiLogger.logState("create table structure...");
	try{
		TransferFromIli trsfFromIli=new TransferFromIli();
		
		// map ili-classes to sql-tables
		// TODO move default SRS to config
		DbSchema schema;
		try {
			schema = trsfFromIli.doit(td,mapping,config.getDefaultSrsAuthority(),config.getDefaultSrsCode());
		} catch (Ili2dbException e) {
			EhiLogger.logError("mapping of ili-classes to sql-tables failed",e);
			return;
		}
		if(schema==null){
			return;
		}

		TransferFromIli.addInheritanceTable(schema);
		Mapping.addMappingTable(schema);
		DbIdGen.addMappingTable(schema);
					
		GeneratorDriver drv=new GeneratorDriver(gen);
		SqlConfig sqlconfig=new SqlConfig();
		sqlconfig.setJdbcConnection(conn);
		drv.visitSchema(sqlconfig,schema);
		// is a create script requested by user?
		String createscript=config.getCreatescript();
		if(createscript!=null && (gen instanceof GeneratorJdbc)){
			writeScript(createscript,((GeneratorJdbc)gen).iteratorCreateLines());
		}
		// is a drop script requested by user?
		String dropscript=config.getDropscript();
		if(dropscript!=null && (gen instanceof GeneratorJdbc)){
			writeScript(dropscript,((GeneratorJdbc)gen).iteratorDropLines());
		}
		// update mapping table
		mapping.updateMappingTable(conn);
		// update inheritance table
		trsfFromIli.updateInheritanceTable(conn);
	}catch(java.io.IOException ex){
		EhiLogger.logError(ex);
		return;
	}
	
	try{
		if(conn!=null){
			conn.close();
			conn=null;
		}
		EhiLogger.logState("...done");
	}catch(java.sql.SQLException ex){
		EhiLogger.logError(ex);
	}
}
	public static void runExport(Config config,String appHome)
	throws Ili2dbException
	{
		String xtffile=config.getXtffile();
		if(xtffile==null){
			EhiLogger.logError("no xtf-file given");
			return;
		}
		String modeldir=config.getModeldir();
		if(modeldir==null){
			EhiLogger.logError("no modeldir given");
			return;
		}
		ArrayList modeldirv = getModelDirv(modeldir, xtffile,appHome);

		String dburl=config.getDburl();
		String dbusr=config.getDbusr();
		String dbpwd=config.getDbpwd();
		if(dburl==null){
			EhiLogger.logError("no dburl given");
			return;
		}
		if(dbusr==null){
			//EhiLogger.logError("no dbusr given");
			//return;
			dbusr="";
		}
		if(dbpwd==null){
			//EhiLogger.logError("no dbpwd given");
			//return;
			dbpwd="";
		}
		String geometryConverter=config.getGeometryConverter();
		if(geometryConverter==null){
			EhiLogger.logError("no geoemtry converter given");
			return;
		}
		String jdbcDriver=config.getJdbcDriver();
		if(jdbcDriver==null){
			EhiLogger.logError("no JDBC driver given");
			return;
		}
		
		String models=config.getModels();
		ArrayList modelv=new ArrayList();
		if(models==null){
			EhiLogger.logError("no models given");
			return;
		}
		String modelnames[]=models.split(";");
		for(int modeli=0;modeli<modelnames.length;modeli++){
			String m=modelnames[modeli];
			if(m.equals(XTF)){
				// TODO read modelname from db
			}
			modelv.add(m);				
		}
		if(modelv.size()==0){
			EhiLogger.logError("no models given");
			return;
		}

		String adapterClassName=config.getGeometryConverter();
		if(adapterClassName==null){
			EhiLogger.logError("no adapter given");
			return;
		}
		
		// compile required ili files
		EhiLogger.logState("compile models...");
		TransferDescription td;
		try {
			if(((String)modelv.get(0)).endsWith(".ili")){
				td = ch.interlis.ili2c.Ili2c.compileIliFiles(modelv,modeldirv);
			}else{
				td = ch.interlis.ili2c.Ili2c.compileIliModels(modelv,modeldirv);				
			}
		} catch (Ili2cFailure ex) {
			EhiLogger.logError(ex);
			return;
		}

		SqlGeometryConverter geomConverter=null;
		try{
			geomConverter=(SqlGeometryConverter)Class.forName(geometryConverter).newInstance();
		}catch(Exception ex){
			EhiLogger.logError("failed to load/create geometry converter",ex);
			return;
		}
		
		// open db connection
		try{
			Class.forName(jdbcDriver);
		}catch(Exception ex){
			EhiLogger.logError("failed to load JDBC driver",ex);
			return;
		}
		Connection conn=null;
		String url = dburl;
		try{
		  //DriverManager.registerDriver(new oracle.jdbc.OracleDriver());
		  conn = DriverManager.getConnection(url, dbusr, dbpwd);

		  geomConverter.setup(conn, config);
		  
		  // get mapping definition
		  String mappingConfig=config.getMappingConfig();
		  Mapping mapping=new Mapping();
		  if(DbUtility.tableExists(conn,Mapping.SQL_T_ILI2DB_CLASSNAME)){
			  // read mapping from db
			  mapping.readMappingTable(conn);
		  }else if(mappingConfig!=null){
			  // read mapping from config file if it doesn't exist in the db
			  mapping.readConfig(mappingConfig);
		  }
		
		  // process xtf files
		  EhiLogger.logState("process data...");
		  transferToXtf(conn,xtffile,mapping,td,geomConverter,config.getSender());

		  EhiLogger.logState("...export done");

		}catch(Exception ex){
			EhiLogger.logError(ex);
		}finally{
			try{
				conn.close();
			}catch(java.sql.SQLException ex){
				EhiLogger.logError(ex);
			}
		}
		
	}
	public static String getModelFromXtf(String filename)
	{
		ch.interlis.iox.StartBasketEvent be=null;
		try{
			ch.interlis.iom_j.xtf.XtfReader ioxReader=new ch.interlis.iom_j.xtf.XtfReader(new java.io.File(filename));
			// get first basket
			ch.interlis.iox.IoxEvent event;
			do{
				event=ioxReader.read();
				if(event instanceof ch.interlis.iox.StartBasketEvent){
					be=(ch.interlis.iox.StartBasketEvent)event;
					break;
				}
			}while(!(event instanceof ch.interlis.iox.EndTransferEvent));
			ioxReader.close();
			ioxReader=null;
		}catch(ch.interlis.iox.IoxException ex){
			EhiLogger.logError("failed to read model from xml file "+filename,ex);
			return null;
		}
		// no baskets?
		if(be==null){
			// no model
			return null;
		}
		String qtopic[]=be.getType().split("\\.");
		String model=qtopic[0];
		//EhiLogger.debug("model from xtf <"+model+">");
		return model;
	}
	public static ArrayList getModelDirv(String modeldir, String xtffile,String ili2dbHome) {
		ArrayList modeldirv=new ArrayList();
		String modeldirs[]=modeldir.split(";");
		for(int modeli=0;modeli<modeldirs.length;modeli++){
			String m=modeldirs[modeli];
			if(m.equals(XTF_DIR)){
				if(xtffile!=null){
					m=new java.io.File(xtffile).getAbsoluteFile().getParentFile().getAbsolutePath();
				}else{
					m=null;
				}
			}else if(m.equals(JAR_DIR)){
				if(ili2dbHome!=null){
					m=ili2dbHome;
				}else{
					m=null;
				}
			}
			if(m!=null && m.length()>0){
				modeldirv.add(m);				
			}
		}
		return modeldirv;
	}
	static private void writeScript(String filename,Iterator linei)
	throws java.io.IOException
	{
		java.io.PrintWriter out=new java.io.PrintWriter(new java.io.BufferedOutputStream(new java.io.FileOutputStream(filename)));
		while(linei.hasNext()){
		  GeneratorJdbc.AbstractStmt stmt=(GeneratorJdbc.AbstractStmt)linei.next();
		  String line=stmt.getLine();
		  if(stmt instanceof GeneratorJdbc.Stmt){
			out.println(line+";");
		  }else{
			out.println("-- "+line);
		  }
		}
		out.close();				
	}

	/** transfer data from xml file to database
	*/
	static private void transferFromXtf(Connection conn,String xtffile,Mapping ili2sqlName,TransferDescription td,
			String dbusr,
			SqlGeometryConverter geomConv,
			String defaultCrsAuthority,
			String defaultSrcCode){	
		XtfReader reader=null;
		try{
			reader=new XtfReader(new java.io.File (xtffile));
			TransferFromXtf trsfr=new TransferFromXtf(ili2sqlName,td,conn,dbusr,geomConv,defaultCrsAuthority,defaultSrcCode);
			trsfr.doit(reader);
		}catch(ch.interlis.iox.IoxException ex){
			EhiLogger.logError("failed to read xml file",ex);
		}finally{
			if(reader!=null){
				try{
					reader.close();
				}catch(ch.interlis.iox.IoxException ex){
					EhiLogger.logError("failed to close xml file",ex);
				}
				reader=null;
			}
		}
	}
	/** transfer data from database to xml file
	*/
	static private void transferToXtf(Connection conn,String xtffile,Mapping ili2sqlName,TransferDescription td,SqlGeometryConverter geomConv,String sender){	

		java.io.File outfile=new java.io.File(xtffile);
		XtfWriter ioxWriter=null;
		try{
			ioxWriter=new XtfWriter(outfile,td);
			TransferToXtf trsfr=new TransferToXtf(ili2sqlName,td,conn,geomConv);
			trsfr.doit(ioxWriter,sender);
			//trsfr.doitJava();
			ioxWriter.flush();
		}catch(ch.interlis.iox.IoxException ex){
			EhiLogger.logError("failed to write xml output",ex);
		}finally{
			if(ioxWriter!=null){
				try{
					ioxWriter.close();
				}catch(ch.interlis.iox.IoxException ex){
					EhiLogger.logError("failed to close xml output",ex);
				}
			}
			ioxWriter=null;
		}
	}
	
	static private HashSet getModelNames(ArrayList modelv){
		HashSet ret=new HashSet();
		Iterator modeli=modelv.iterator();
		while(modeli.hasNext()){
			String topic=(String)modeli.next(); // modelname or qualified topicname
			int endModelName=topic.indexOf('.');
			if(endModelName<=0){
				// just a modelname
				ret.add(topic);
			}else{
				// qualified topicname; get model name
				String model=topic.substring(0,endModelName);
				ret.add(model);
			}
		}
		return ret;
	}
}
