/* This file is part of the ili2ora project.
 * For more information, please see <http://www.interlis.ch>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package ch.ehi.ili2db.fromili;

import ch.ehi.basics.logging.EhiLogger;

import ch.interlis.ili2c.metamodel.*;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import ch.ehi.sqlgen.repository.*;
import ch.ehi.ili2db.base.Ili2dbException;

/**
 * @author ce
 * @version $Revision: 1.0 $ $Date: 07.02.2005 $
 */
public class TransferFromIli {
	private DbSchema schema=null;
	private HashSet visitedElements=null;
	private TransferDescription td=null;
	private ch.ehi.ili2db.mapping.Mapping ili2sqlName=null;
	private String defaultCrsAuthority=null;
	private String defaultSrcCode=null;
	
	public DbSchema doit(TransferDescription td1,ch.ehi.ili2db.mapping.Mapping ili2sqlName,
		String defaultSrsAuthority,
		String defaultSrsCode)
	throws Ili2dbException
	{
		this.defaultCrsAuthority=defaultSrsAuthority;
		this.defaultSrcCode=defaultSrsCode;
		this.ili2sqlName=ili2sqlName;
		schema=new DbSchema();
		schema.setName("ili2ora");
		visitedElements=new HashSet();
		td=td1;
		Iterator modeli=td.iterator();
		while(modeli.hasNext()){
			Object modelo=modeli.next();
			if(modelo instanceof Model){
				Model model=(Model)modelo;
				Iterator topici=model.iterator();
				while(topici.hasNext()){
					Object tObj=topici.next();
					if (tObj instanceof Topic){
						Topic topic=(Topic)tObj;
						visitTopic(topic);
					}
				}
				visitViewableOfModel(model);
			}
		}
		
		// sys_interlisnames
		// interlis LONGVARCHAR(767)
		// db VARCHAR(30)
		return schema;		

	}
	private void visitViewableOfModel(Model def)
	throws Ili2dbException
	{
		if(visitedElements.contains(def)){
			return;
		}
		visitedElements.add(def);
		Iterator classi=def.iterator();
		while(classi.hasNext()){
			Object classo=classi.next();
			if(classo instanceof Viewable){
				visitViewable((Viewable)classo);
			}
		}
	}
	private void visitTopic(Topic def)
	throws Ili2dbException
	{
		if(visitedElements.contains(def)){
			return;
		}
		visitedElements.add(def);
		Iterator classi=def.iterator();
		while(classi.hasNext()){
			Object classo=classi.next();
			if(classo instanceof Viewable){
				visitViewable((Viewable)classo);
			}
		}
		// base topic
		Topic base=(Topic)def.getExtending();
		if(base!=null){
			visitTopic(base);
		}
		// parent model
		Model model=(Model)def.getContainer(Model.class);
		if(!visitedElements.contains(model)){
			//EhiLogger.debug("new model "+def);
		}
		visitViewableOfModel(model);
	}
	private void visitViewable(Viewable def)
	throws Ili2dbException
	{
		if(visitedElements.contains(def)){
			return;
		}
		visitedElements.add(def);
		try{
			generateViewable(def);
		}catch(Exception ex){
			throw new Ili2dbException(def.getScopedName(null),ex);
		}
		// base viewable
		Viewable base=(Viewable)def.getExtending();
		if(base!=null){
			visitViewable(base);
		}
		// parent model
		Model model=(Model)def.getContainer(Model.class);
		if(!visitedElements.contains(model)){
			//EhiLogger.debug("new model "+def);
		}
		visitViewableOfModel(model);
	}
	private void generateViewable(Viewable def)
	throws Ili2dbException
	{
		if(def instanceof AssociationDef){
			AssociationDef assoc=(AssociationDef)def;
			if(assoc.getDerivedFrom()!=null){
				return;
			}
			if(assoc.isLightweight() 
				&& !assoc.getAttributes().hasNext()
				&& !assoc.getLightweightAssociations().iterator().hasNext()) {
				return;
			}
		}
		//EhiLogger.debug("viewable "+def);
		String sqlName=getSqlTableName(def);
		Viewable base=(Viewable)def.getExtending();
		DbTable dbTable=new DbTable();
		dbTable.setName(sqlName);
		dbTable.setIliName(def.getScopedName(null));
		if(base==null){
		  dbTable.setRequiresSequence(true);
		}
		String baseRef="";
		DbColId dbColId=addKeyCol(dbTable);
		if(base!=null){
		  dbColId.setScriptComment("REFERENCES "+base.getScopedName(null));
		}
		DbColumn dbCol;
		if(base==null){
		  dbCol=createSqlTypeCol("T_Type");
		  dbTable.addColumn(dbCol);
		  // if STRUCTURE, add ref to parent
		  if((def instanceof Table) && !((Table)def).isIdentifiable()){
			  // add parentid
			DbColId dbParentId=new DbColId();
			dbParentId.setName("T_ParentId");
			dbParentId.setNotNull(true);
			dbParentId.setPrimaryKey(false);
			dbTable.addColumn(dbParentId);
			  // add parent_type
			dbCol=createSqlTypeCol("T_ParentType");
			dbTable.addColumn(dbCol);
			// add parent_attr
			dbCol=createSqlTypeCol("T_ParentAttr");
			dbTable.addColumn(dbCol);
			// add seqeunce attr
			DbColId dbSeq=new DbColId();
			dbSeq.setName("T_Seq");
			dbSeq.setNotNull(true);
			dbSeq.setPrimaryKey(false);
			dbTable.addColumn(dbSeq);
		  }
		}

		// body
		Iterator iter=def.getDefinedAttributesAndRoles2();
		  while (iter.hasNext()) {
			  ViewableTransferElement obj = (ViewableTransferElement)iter.next();
			  if (obj.obj instanceof AttributeDef) {
				  AttributeDef attr = (AttributeDef) obj.obj;
				  if(attr.getExtending()==null){
					try{
						dbCol=generateType(attr);
					}catch(Exception ex){
						throw new Ili2dbException(attr.getScopedName(null),ex);
					}
					if(dbCol!=null){
					  dbCol.setName(getSqlAttrName(attr));
					  if(attr.getDomain().isMandatoryConsideringAliases()){
						dbCol.setNotNull(true);
					  }
					  dbTable.addColumn(dbCol);
					}
				  }
			  }
			  if(obj.obj instanceof RoleDef){
				  RoleDef role = (RoleDef) obj.obj;
					if(role.getExtending()==null){
						// not an embedded role and roledef not defined in a lightweight association?
						if (!obj.embedded && !((AssociationDef)def).isLightweight()){
						  dbColId=new DbColId();
						  dbColId.setName(getSqlRoleName(role));
						  dbColId.setNotNull(true);
						  dbColId.setPrimaryKey(false);
						  dbTable.addColumn(dbColId);
						  // TODO handle ordered
						}
						// a role of an embedded association?
						if(obj.embedded){
							AssociationDef roleOwner = (AssociationDef) role.getContainer();
							if(roleOwner.getDerivedFrom()==null){
								// role is oppend;
							  dbColId=new DbColId();
							  dbColId.setName(getSqlRoleName(role));
							  dbColId.setNotNull(false);
							  dbColId.setPrimaryKey(false);
							  dbTable.addColumn(dbColId);
							  // TODO handle ordered
							}
						}
					}
			  }
		}
		addStdCol(dbTable);
	  	schema.addTable(dbTable);
	}
	static public boolean isBoolean(TransferDescription td,Type type){
		while(type instanceof TypeAlias) {
			if (((TypeAlias) type).getAliasing() == td.INTERLIS.BOOLEAN) {
				return true;
			}
			type=((TypeAlias) type).getAliasing().getType();
		}
		
		return false;
	}
	private DbColumn generateType(AttributeDef attr)
	throws Ili2dbException
	{
		Type type = attr.getDomain();
		if (type instanceof TypeAlias) {
			if (isBoolean(td,type)) {
				return new DbColBoolean();
			}
		}
		type = attr.getDomainResolvingAliases();
		if (type instanceof PolylineType){
			DbColGeometry ret=new DbColGeometry();
			ret.setType(DbColGeometry.LINESTRING);
			// TODO get crs from ili
			ret.setSrsAuth(defaultCrsAuthority);
			ret.setSrsId(defaultSrcCode);
			Domain coordDomain=((PolylineType)type).getControlPointDomain();
			if(coordDomain!=null){
				CoordType coord=(CoordType)coordDomain.getType();
				ret.setDimension(coord.getDimensions().length);
				setBB(ret, coord,attr);
			}
			return ret;
		}else if (type instanceof SurfaceOrAreaType){
			DbColGeometry ret=new DbColGeometry();
			ret.setType(DbColGeometry.POLYGON);
			// TODO get crs from ili
			ret.setSrsAuth(defaultCrsAuthority);
			ret.setSrsId(defaultSrcCode);
			CoordType coord=(CoordType)((SurfaceOrAreaType)type).getControlPointDomain().getType();
			ret.setDimension(coord.getDimensions().length);
			setBB(ret, coord,attr);
			return ret;
		}else if (type instanceof CoordType){
			DbColGeometry ret=new DbColGeometry();
			ret.setType(DbColGeometry.POINT);
			// TODO get crs from ili
			ret.setSrsAuth(defaultCrsAuthority);
			ret.setSrsId(defaultSrcCode);
			CoordType coord=(CoordType)type;
			ret.setDimension(coord.getDimensions().length);
			setBB(ret, coord,attr);
			return ret;
		}else if (type instanceof CompositionType){
			// skip it
			visitViewable(((CompositionType)type).getComponentType());
			return null;
		}else if (type instanceof ReferenceType){
			DbColId ret=new DbColId();
			ret.setNotNull(false);
			ret.setPrimaryKey(false);
			return ret;
		}else if (type instanceof BasketType){
			// skip it; type no longer exists in ili 2.3
			return null;
		//}else if(type instanceof EnumerationType){
		}else if(type instanceof NumericType){
			if(type.isAbstract()){
			}else{
				PrecisionDecimal min=((NumericType)type).getMinimum();
				PrecisionDecimal max=((NumericType)type).getMaximum();
				if(min.getAccuracy()>0){
					DbColDecimal ret=new DbColDecimal();
					int size=Math.max(min.toString().length(),max.toString().length());
					int precision=min.getAccuracy();
					//EhiLogger.debug("attr "+ attr.getName()+", maxStr <"+maxStr+">, size "+Integer.toString(size)+", precision "+Integer.toString(precision));
					ret.setSize(size);
					ret.setPrecision(precision);
					return ret;
				}else{
					DbColNumber ret=new DbColNumber();
					int size=Math.max(min.toString().length(),max.toString().length());
					ret.setSize(size);
					return ret;
				}
			}
		}else if(type instanceof TextType){
			DbColVarchar ret=new DbColVarchar();
			if(((TextType)type).getMaxLength()>0){
				ret.setSize(((TextType)type).getMaxLength());
			}else{
				ret.setSize(255);
			}
			return ret;
		}

		DbColVarchar ret=new DbColVarchar();
		ret.setSize(255);
		return ret;
		
	}
	private void setBB(DbColGeometry ret, CoordType coord,AttributeDef attr) {
		NumericalType dimv[]=coord.getDimensions();
		if(!(dimv[0] instanceof NumericType) || !(dimv[1] instanceof NumericType)){
			EhiLogger.logError("Attribute "+attr.getScopedName(null)+": COORD type not supported ("+dimv[0].getClass().getName()+")");
			return;
		}
		if(((NumericType)dimv[0]).getMinimum()!=null){
			ret.setMin1(((NumericType)dimv[0]).getMinimum().doubleValue());
			ret.setMax1(((NumericType)dimv[0]).getMaximum().doubleValue());
			ret.setMin2(((NumericType)dimv[1]).getMinimum().doubleValue());
			ret.setMax2(((NumericType)dimv[1]).getMaximum().doubleValue());
			if(dimv.length==3){
				ret.setMin3(((NumericType)dimv[2]).getMinimum().doubleValue());
				ret.setMax3(((NumericType)dimv[2]).getMaximum().doubleValue());
			}
		}
	}
	/** setup a mapping from a qualified model or topic name
	 * to the corresponding java object.
	 */
	private HashMap setupTopicTagMap(TransferDescription td){
		HashMap ret=new HashMap();
		Iterator modeli = td.iterator ();
		while (modeli.hasNext ())
		{
		  Object mObj = modeli.next ();
		  if(mObj instanceof Model){
			Model model=(Model)mObj;
			ret.put(model.getScopedName(null),model);
			Iterator topici=model.iterator();
			while(topici.hasNext()){
				Object tObj=topici.next();
				if (tObj instanceof Topic){
				  	Topic topic=(Topic)tObj;
   					ret.put(topic.getScopedName(null),topic);
				}
			}
		  }
		}
		return ret;
	}
	private String getSqlRoleName(RoleDef def){
		return ili2sqlName.mapIliRoleDef(def);
	}
	private String getSqlAttrName(AttributeDef def){
		return ili2sqlName.mapIliAttributeDef(def);
	}
	private String getSqlTableName(Viewable def){
		String sqlname=ili2sqlName.mapIliClassDef(def);
		return sqlname;
	}
	private DbColumn createSqlTypeCol(String name){
	  DbColVarchar dbCol=new DbColVarchar();
	  dbCol.setName(name);
	  dbCol.setNotNull(true);
	  dbCol.setSize(30);
	  return dbCol;
	}
	private DbColId addKeyCol(DbTable table) {
		  DbColId dbColId=new DbColId();
		  dbColId.setName("T_Id");
		  dbColId.setNotNull(true);
		  dbColId.setPrimaryKey(true);
		  table.addColumn(dbColId);
		  return dbColId;
	}
	public static void addStdCol(DbTable table) {
		DbColumn dbCol=new DbColDateTime();
		dbCol.setName("T_LastChange");
		dbCol.setNotNull(true);
		table.addColumn(dbCol);
	
		dbCol=new DbColDateTime();
		dbCol.setName("T_CreateDate");
		dbCol.setNotNull(true);
		table.addColumn(dbCol);
	
		DbColVarchar dbColUsr=new DbColVarchar();
		dbColUsr.setName("T_User");
		dbColUsr.setNotNull(true);
		dbColUsr.setSize(40);
		table.addColumn(dbColUsr);
	}
	public static final String INHERIT_TAB="T_ILI2DB_INHERITANCE";
	public static final String INHERIT_TAB_THIS="thisClass";
	public static final String INHERIT_TAB_BASE="baseClass";
	static public void addInheritanceTable(DbSchema schema)
	{
		DbTable tab=new DbTable();
		tab.setName(INHERIT_TAB);
		DbColVarchar thisClass=new DbColVarchar();
		thisClass.setName(INHERIT_TAB_THIS);
		thisClass.setNotNull(true);
		thisClass.setSize(30);
		tab.addColumn(thisClass);
		DbColVarchar baseClass=new DbColVarchar();
		baseClass.setName(INHERIT_TAB_BASE);
		baseClass.setNotNull(false);
		baseClass.setSize(30);
		tab.addColumn(baseClass);
		schema.addTable(tab);
	}
	public void updateInheritanceTable(java.sql.Connection conn)
	throws Ili2dbException
	{
		//String stmt="CREATE TABLE "+tabname+" ("+thisClassCol+" VARCHAR2(30) NOT NULL,"+baseClassCol+" VARCHAR2(30) NULL)";
		try{

			// insert entries
			String stmt="INSERT INTO "+INHERIT_TAB+" ("+INHERIT_TAB_THIS+","+INHERIT_TAB_BASE+") VALUES (?,?)";
			java.sql.PreparedStatement ps = conn.prepareStatement(stmt);
			String thisClass=null;
			try{
				java.util.Iterator entri=visitedElements.iterator();
				while(entri.hasNext()){
					Object entro=entri.next();
					if(entro instanceof Viewable){
						Viewable aclass=(Viewable)entro;
						thisClass=getSqlTableName(aclass);
						Viewable base=(Viewable)aclass.getExtending();
						ps.setString(1, thisClass);
						if(base!=null){
							String baseClass=getSqlTableName(base);
							ps.setString(2, baseClass);
						}else{
							ps.setNull(2,java.sql.Types.VARCHAR);
						}
						ps.executeUpdate();
					}
				}
			}catch(java.sql.SQLException ex){
				throw new Ili2dbException("failed to insert inheritance-relation for class "+thisClass,ex);
			}finally{
				ps.close();
			}
		}catch(java.sql.SQLException ex){		
			throw new Ili2dbException("failed to update inheritance-table "+INHERIT_TAB,ex);
		}

	}
}
